#!/bin/env python
#Copyright ReportLab Europe Ltd. 2000-2017
#see license.txt for license details
#history https://bitbucket.org/rptlab/reportlab/history-node/tip/src/reportlab/lib/__init__.py
__version__='3.3.0'
import os
RL_DEBUG = 'RL_DEBUG' in os.environ
